/* -*- c-basic-offset: 4; -*-
 * Copyright (c) 2005 Hans Insulander
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main.cc,v 1.10 2005/11/10 14:27:13 hin Exp $
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>

#include <GL/glfw.h>

#include "common.hh"

const char *title = "Spectrum Analyser";

int width = 400;
int height = 200;

double ref_level = 1.0; // Reference level, typically max input level
double min_level = -60; // smallest level in dB

void
set_viewport() {
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(-.1, 1.1, -.1, 1.1);
    glMatrixMode(GL_MODELVIEW);
}

void
fps() {
    static double lasttime = 0;
    static int frames = 0;
    frames++;
    double now = glfwGetTime();
    if(lasttime == 0 || now >= (lasttime + 1.0)) {
	double fps = (double)frames / (now - lasttime);
	char s[128];
	snprintf(s, sizeof s, "%s: %.0f fps", title, fps);
	glfwSetWindowTitle(s);

	frames = 0;
	lasttime = now;
    }
}

double bands[] = { 20, 25, 31.5,
		   40, 50, 63,
		   80, 100, 125,
		   160, 200, 250,
		   315, 400, 500,
		   630, 800, 1000,
		   1250, 1600, 2000,
		   2500, 3150, 4000,
		   5000, 6300, 8000,
		   10000, 12500, 16000,
		   20000 };
int num_bands = sizeof(bands) / sizeof(bands[0]);

double min_freq = bands[0];
double max_freq = bands[num_bands-1];

double
fscale(double f) {
#ifdef FLINEAR
    return (f-min_freq)/(max_freq-min_freq);
#else
    return (log(f)-log(min_freq))/(log(max_freq)-log(min_freq));
#endif
}

double dB(double lin) {
    return 10*log(lin/ref_level)/log(10);
}

double
ascale(double level) {
    return 1.0-(dB(level)/min_level);
}

void GLFWCALL
resize(int w, int h) {
    width = w;
    height = h;
    set_viewport();
}

int
main(int argc, char **argv) {
    int running = GL_TRUE;
    bool window_mode = true;
    int window_width, window_height;

    glfwInit();
    glfwSwapInterval(1);

    audio_init();

    if(!glfwOpenWindow(width, height, 0, 0, 0, 0, 0, 0, GLFW_WINDOW)) {
	glfwTerminate();
	return 0;
    }

    glEnable(GL_DEPTH_TEST);

    set_viewport();
    glfwSetWindowSizeCallback(resize);

    float re[FFTLEN/2];
    float im[FFTLEN/2];

    while(running) {
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	if(fftdata_re != NULL && fftdata_im != NULL) {
	    memcpy(re, (const void *)fftdata_re, sizeof(re));
	    memcpy(im, (const void *)fftdata_im, sizeof(im));
	    fftdata_re = NULL;
	    fftdata_im = NULL;
	}

	/* 
	 * Draw the FFT analyzed data
	 */
	glColor3f(1, 1, 0); /* yellow */
	glBegin(GL_LINE_STRIP);
	for(int i=1; i<FFTLEN/2; i++) {
	    double f = ((double)i/(double)(FFTLEN/2))*SAMPLERATE/2;
	    double a = sqrt(re[i]*re[i] + im[i]*im[i]);
	    glVertex2f(fscale(f), ascale(a));
	};
	glEnd();

	/*
	 * Draw ISO 1/3 octave band center frequencies
	 */
	glColor3f(0, 0, 1); /* blue */
	glBegin(GL_LINES);
	for(int i=0; i<num_bands; i++) {
	    double f = bands[i];
	    glVertex2f(fscale(f), 0);
	    glVertex2f(fscale(f), -0.03);
	};
	glEnd();

	/*
	 * Draw amplitude scale, 10dB marks
	 */
	glColor3f(1, 0, 0); /* red */
	glBegin(GL_LINES);
	for(double a = ref_level; ascale(a) >= 0.0; a /= 10) {
	    glVertex2f(-0.03, ascale(a));
	    glVertex2f(0, ascale(a));
	}
	glEnd();

	/*
	 * Draw a nice border
	 */
	glColor3f(0, 0.5, 0); /* green */
	glBegin(GL_LINE_STRIP);
	glVertex2f(0, ascale(ref_level));
	glVertex2f(1, ascale(ref_level));
	glVertex2f(1, 0);
	glVertex2f(0, 0);
	glVertex2f(0, ascale(ref_level));
	glEnd();

	glfwSwapBuffers();
	running = !glfwGetKey(GLFW_KEY_ESC) &&
	    glfwGetWindowParam(GLFW_OPENED);

	/*
	 * Switch to fullscreen
	 */
	if(glfwGetKey('F') && window_mode) {
	    /* remember window size so we can restore it */
	    window_width = width;
	    window_height = height;
	    
	    GLFWvidmode mode;
	    glfwGetDesktopMode(&mode);
	    glfwCloseWindow();
	    glfwOpenWindow(mode.Width, mode.Height,
			   mode.RedBits, mode.GreenBits, mode.BlueBits,
			   0, 0, 0, GLFW_FULLSCREEN);
	    resize(mode.Width, mode.Height);
	    window_mode = false;
	}

	/*
	 * Switch to window mode
	 */
	if(glfwGetKey('W') && !window_mode) {
	    glfwCloseWindow();
	    glfwOpenWindow(window_width, window_height,
			   0, 0, 0,
			   0, 0, 0, GLFW_WINDOW);
	    resize(window_width, window_height);
	    window_mode = true;
	}

	fps();
    }

    glfwTerminate();
    return 0;
}
